@@ -1,10 +1,17 @@
 #!/usr/bin/env python3
 
+import asyncio
+import concurrent.futures
+import datetime
 import hashlib
 import json
+import time
+import re
 import os
 import secrets
+import time
 import urllib.request
+import urllib.error
 
 from decimal import Decimal
 from typing import Tuple
@@ -26,6 +33,7 @@ from chia.wallet.puzzles.p2_delegated_puzzle_or_hidden_puzzle import (
 from chia.wallet.puzzles.load_clvm import load_clvm
 
 from .bech32m import bech32_decode, convertbits
+from .pushtx import push_tx
 
 
 GROUP_ORDER = 0x73EDA753299D7D483339D80809A1D80553BDA402FFFE5BFEFFFFFFFF00000001
@@ -40,7 +48,6 @@ OWNERSHIP_MESSAGE = b"I own this key"
 
 
 def lookup_xch_prices():
-    print("prices from cryptocompare.com")
     prices = json.load(
         urllib.request.urlopen(
             "https://min-api.cryptocompare.com/data/price?fsym=XCH&tsyms=USD,BTC"
@@ -53,14 +60,57 @@ def lookup_xch_prices():
 def fromhex(s):
     if s.lower().startswith("0x"):
         s = s[2:]
+    if len(s) & 1 == 1:
+        s = f"0{s}"
     return bytes.fromhex(s)
 
 
+def wait_for_payments_to_address(address, min_amount):
+    root_url = "https://api2.spacescan.io"
+    addr_url = f"{root_url}/1/xch/address/txns"
+    while 1:
+        print(f"checking for payments to {address}")
+        resp = urllib.request.urlopen(f"{addr_url}/{address}", timeout=10)
+        if resp.status == 200:
+            coins = json.load(resp, parse_float=Decimal)["data"]["coins"][::-1]
+            for coin in coins:
+                amount = int(coin["amount"])
+                if amount >= min_amount:
+                    return [coin]
+        time.sleep(5)
+
+
 # ### ui
 
 
+def ui_choose_path(possible_paths):
+    if len(possible_paths) > 0:
+        while 1:
+            print(f" 0. NEW SWAP")
+            for idx, pp in enumerate(possible_paths):
+                print(f"{(idx+1):2}. {pp}")
+            r = input("> ")
+            try:
+                v = int(r)
+                if v == 0:
+                    break
+                return possible_paths[v - 1]
+            except ValueError:
+                pass
+    name = input('(optional) short name for counterparty (example: "ed")> ')
+    if len(name) > 0:
+        name = f"-{name}"
+    now = datetime.datetime.now().strftime("%Y-%m-%d-%H%M%S")
+    path = f"xchswap-log-{now}{name}.txt"
+    return path
+
+
 def ui_get_logfile():
-    path = "xchlog.txt"
+    cre = re.compile(r"xchswap-log-(\d{4})-(\d{2})-(\d{2})-(\d{6})(\-.+)?.txt")
+    possible_paths = [_ for _ in os.listdir(".") if cre.match(_)]
+    possible_paths.sort()
+    path = ui_choose_path(possible_paths)
+
     if not os.path.exists(path):
         with open(path, "a") as f:
             secret_key = secrets.randbits(256)
@@ -76,8 +126,6 @@ def ui_get_logfile():
             line, file_lines = file_lines[0], file_lines[1:]
             if line.startswith("#"):
                 continue
-            import time
-
             time.sleep(0.125)
             print("%s%s" % (msg, line))
             return line[:-1]
@@ -98,8 +146,8 @@ def ui_choose(input):
             return int(choice)
 
 
-def ui_get_amounts(input):
-    prices = lookup_xch_prices()
+def ui_get_amounts(input, prices):
+    print("prices from cryptocompare.com")
     BTC_PER_XCH = prices["BTC"]
     USD_PER_XCH = prices["USD"]
     USD_PER_BTC = USD_PER_XCH / BTC_PER_XCH
@@ -114,7 +162,7 @@ def ui_get_amounts(input):
     xch_amount = Decimal(input("How much XCH is being traded? > "))
     btc_amount = xch_amount * BTC_PER_XCH
     print(
-        "%s XCH worth about %0.6f btc (USD$%0.2f)"
+        "%0.13f XCH worth about %0.8f btc (USD$%0.2f)"
         % (xch_amount, btc_amount, USD_PER_XCH * xch_amount)
     )
     print()
@@ -138,30 +186,6 @@ def ui_get_lightning_payment_request(input):
         print("invalid")
 
 
-def ui_get_parent_coin_id(input):
-    while 1:
-        r = input("parent id> ")
-        try:
-            b = fromhex(r)
-            if len(b) == 32:
-                return b
-        except Exception:
-            pass
-        print("parent coin id is a 64 character hex string")
-
-
-def ui_get_pubkey(input):
-    while 1:
-        r = input("enter counter-party pubkey> ")
-        try:
-            b = fromhex(r)
-            g1 = blspy.G1Element.from_bytes(b)
-            return g1
-        except Exception:
-            pass
-        print("pubkey is a 96 character hex string")
-
-
 def ui_get_private_key(input, public_key):
     while 1:
         r = input("enter counter-party private key\n> ")
@@ -177,7 +201,7 @@ def ui_get_private_key(input, public_key):
         print("private key is a 64 character hex string")
 
 
-def ui_get_pubkey_with_sig(input):
+def ui_get_pubkey_with_sig(input, my_pubkey):
     while 1:
         r = input("> ")
         r = r.strip()
@@ -185,6 +209,10 @@ def ui_get_pubkey_with_sig(input):
             puzzle_hash_hex, sig_hex = r.split("_")
             b1 = fromhex(puzzle_hash_hex)
             g1 = blspy.G1Element.from_bytes(b1)
+            if g1 == my_pubkey:
+                print()
+                print("that's your public key, silly! Try again.")
+                continue
             b2 = fromhex(sig_hex)
             g2 = blspy.G2Element.from_bytes(b2)
             r = blspy.AugSchemeMPL.verify(g1, OWNERSHIP_MESSAGE, g2)
@@ -212,6 +240,32 @@ def ui_get_sweep_preimage(input, sweep_receipt_hash):
         print("invalid, try again")
 
 
+def ui_get_sweep_preimage_or_private_key(
+    input, sweep_receipt_hash, clawback_public_key
+):
+    while 1:
+        r = input("> ").strip().lower()
+        if r == "quit":
+            return None, None, True
+        try:
+            p = int(r, 16) % GROUP_ORDER
+            private_key = private_key_for_secret(p)
+            if private_key.get_g1() == clawback_public_key:
+                return None, p, False
+            b = fromhex(r)
+            if hashlib.sha256(b).digest() == sweep_receipt_hash:
+                return b, None, False
+        except Exception:
+            pass
+        print("this isn't the private key nor the pre-image")
+        print("each is a 64 character hex string")
+
+
+def ui_should_send_spend_bundle():
+    r = input("send this spend bundle? (Y/N) > ")
+    return r.lower().startswith("y")
+
+
 # ### end ui
 
 
@@ -390,8 +444,7 @@ def have_xch_want_btc(logfile, secret_key, btc_amount, xch_amount_mojos):
     print()
 
     print("enter your counter-party's public key as pasted by them")
-    sweep_public_key = ui_get_pubkey_with_sig(logfile)
-    print()
+    sweep_public_key = ui_get_pubkey_with_sig(logfile, clawback_public_key)
 
     total_pubkey = sweep_public_key + clawback_public_key
 
@@ -418,26 +471,18 @@ def have_xch_want_btc(logfile, secret_key, btc_amount, xch_amount_mojos):
     puzzle_hash = puzzle_reveal.get_tree_hash()
     address = encode_puzzle_hash(puzzle_hash, "xch")
     xch_amount = Decimal(xch_amount_mojos) / Decimal(int(1e12))
-    print()
+
     print(f"go into your XCH wallet and send {xch_amount} XCH to")
     print(f"{address}")
     print()
-    print("Go to an explorer and look up the parent coin info/name (32 byte hex)")
-    print(f"https://chia.tt/info/address/{address}")
-    print(" => then click on `Coin Name`")
-    # print(f"https://xchscan.com/address/{address} (not sure how to find it here)")
-    # print(f"https://www.chiaexplorer.com/blockchain/address/{address} (not sure how to find it here)")
-    print()
-    print("Enter it below")
-    parent_coin_id = ui_get_parent_coin_id(logfile)
 
-    print()
     print("You need to enter a refund address where your XCH will be returns if")
     print("the swap fails. It can be an address from a wallet or an exchange.")
     print()
-
     clawback_puzzle_hash = ui_get_puzzle_hash(logfile, "enter XCH refund address > ")
 
+    coins = wait_for_payments_to_address(address, xch_amount_mojos)
+    parent_coin_id = fromhex(coins[0]["coin_parent"])
     conditions = [[51, clawback_puzzle_hash, xch_amount_mojos]]
 
     spend_bundle = generate_spendbundle(
@@ -470,8 +515,31 @@ def have_xch_want_btc(logfile, secret_key, btc_amount, xch_amount_mojos):
     print(f"after {hours} hours.")
     print()
     print(f"clawback spend bundle: {spend_bundle_hex}")
-    # spend_bundle.debug()
-    return
+    print()
+    print(f"waiting {clawback_delay_seconds} s then pushing the clawback spend bundle")
+    print(f"Leave this window open or control-c to exit.")
+    print()
+    print(
+        f"Warning: if you answer before {clawback_delay_seconds} seconds have elapsed,"
+    )
+    print(
+        "the spend bundle will be rejected. No harm done though, you'll just have to try later."
+    )
+    if ui_should_send_spend_bundle():
+        try_to_push_tx(spend_bundle, clawback_puzzle_hash)
+
+
+def try_to_push_tx(sb, dest_puzzle_hash):
+    print()
+    print(f"Check your wallet or an explorer to confirm.")
+    address = encode_puzzle_hash(dest_puzzle_hash, "xch")
+    print(f"https://www.spacescan.io/xch/address/{address}")
+    print()
+    r = asyncio.run(push_tx(sb))
+    if r == 0:
+        print("It seems to have worked.")
+    else:
+        print("*** The spend bundle may not have been accepted.")
 
 
 def have_btc_want_xch(logfile, secret_key, btc_amount, xch_amount_mojos):
@@ -484,15 +552,12 @@ def have_btc_want_xch(logfile, secret_key, btc_amount, xch_amount_mojos):
     print()
 
     print("enter your counter-party's public key as pasted by them")
-    clawback_public_key = ui_get_pubkey_with_sig(logfile)
-    print()
+    clawback_public_key = ui_get_pubkey_with_sig(logfile, sweep_public_key)
 
     total_pubkey = sweep_public_key + clawback_public_key
 
-    print()
     print("Paste the lightning payment request from your counter-party here.")
     lpr = ui_get_lightning_payment_request(logfile)
-    print()
     d = parse_lpr(lpr)
     sweep_receipt_hash = d[1]
 
@@ -509,19 +574,22 @@ def have_btc_want_xch(logfile, secret_key, btc_amount, xch_amount_mojos):
     puzzle_hash = puzzle_reveal.get_tree_hash()
     address = encode_puzzle_hash(puzzle_hash, "xch")
     xch_amount = Decimal(xch_amount_mojos) / Decimal(int(1e12))
-    print(f"Your counter-party should be sending {xch_amount} XCH to the address")
+
+    print("Enter an address where your XCH will be delivered.")
+    print("It can be an address from a wallet or an exchange.")
+    print()
+    sweep_puzzle_hash = ui_get_puzzle_hash(logfile, "XCH address > ")
+
+    print(f"Your counter-party should be sending {xch_amount:.13f} XCH to the address")
     print(f"{address}")
     print()
-    print("Go to an explorer and look up the parent coin info/name (32 byte hex)")
-    print(f"https://chia.tt/info/address/{address}")
-    print(" => then click on `Coin Name`")
-    # print(f"https://xchscan.com/address/{address} (not sure how to find it here)")
-    # print(
-    #    f"https://www.chiaexplorer.com/blockchain/address/{address} (not sure how to find it here)"
-    # )
+    print("Go to an explorer and watch for payments")
     print()
-    print("Enter it below")
-    parent_coin_id = ui_get_parent_coin_id(logfile)
+    print(f"https://www.spacescan.io/xch/address/{address}")
+    print()
+    coins = wait_for_payments_to_address(address, xch_amount_mojos)
+    parent_coin_id = fromhex(coins[0]["coin_parent"])
+
     print()
     print("Once your XCH has enough confirmations, pay the lightning invoice.")
     print()
@@ -533,14 +601,9 @@ def have_btc_want_xch(logfile, secret_key, btc_amount, xch_amount_mojos):
     print(f"private key: 0x{s:064x}")
     print()
 
-    print("Enter the lightning invoice receipt preimage")
-    sweep_preimage = ui_get_sweep_preimage(logfile, sweep_receipt_hash)
-
-    print()
-    print("Enter an address where your XCH will be delivered.")
-    print("It can be an address from a wallet or an exchange.")
-    print()
-    sweep_puzzle_hash = ui_get_puzzle_hash(logfile, "XCH address > ")
+    print("Once you've paid the lightning invoice, ask your counterparty to")
+    print("share their private key. Meanwhile, look up your lightning invoice")
+    print("receipt pre-image in case your counterparty doesn't respond.")
     print()
 
     # TODO: fix these next two lines
@@ -554,6 +617,78 @@ def have_btc_want_xch(logfile, secret_key, btc_amount, xch_amount_mojos):
         + ADDITIONAL_DATA
     )
 
+    while True:
+        print(
+            "Enter your counterparty private key OR the lightning invoice receipt pre-image or `quit`"
+        )
+        (
+            sweep_preimage,
+            remote_secret,
+            should_quit,
+        ) = ui_get_sweep_preimage_or_private_key(
+            logfile, sweep_receipt_hash, clawback_public_key
+        )
+
+        if should_quit:
+            break
+
+        if sweep_preimage:
+            sweep_spend_bundle = handle_sweep_preimage(
+                s,
+                puzzle_hash,
+                parent_coin_id,
+                xch_amount_mojos,
+                total_pubkey,
+                clawback_delay_seconds,
+                clawback_public_key,
+                sweep_receipt_hash,
+                sweep_public_key,
+                conditions,
+                sweep_preimage,
+            )
+            print("You should wait for your counterparty to send their private key")
+            print("and only use this spend bundle if they seem non-responsive.")
+            print(
+                f"Warning: after {clawback_delay_seconds} s they can claw back the XCH"
+            )
+            print()
+            if ui_should_send_spend_bundle():
+                try_to_push_tx(sweep_spend_bundle, sweep_puzzle_hash)
+
+        if remote_secret:
+            clean_spend_bundle = handle_remote_secret(
+                coin_spend,
+                message,
+                remote_secret,
+                s,
+                synthetic_offset,
+                parent_coin_id,
+                xch_amount_mojos,
+                total_pubkey,
+                clawback_delay_seconds,
+                clawback_public_key,
+                sweep_receipt_hash,
+                sweep_public_key,
+                conditions,
+            )
+            print()
+            if ui_should_send_spend_bundle():
+                try_to_push_tx(clean_spend_bundle, sweep_puzzle_hash)
+
+
+def handle_sweep_preimage(
+    my_secret,
+    puzzle_hash,
+    parent_coin_id,
+    xch_amount_mojos,
+    total_pubkey,
+    clawback_delay_seconds,
+    clawback_public_key,
+    sweep_receipt_hash,
+    sweep_public_key,
+    conditions,
+    sweep_preimage,
+):
     spend_bundle = generate_spendbundle(
         parent_coin_id,
         xch_amount_mojos,
@@ -573,43 +708,62 @@ def have_btc_want_xch(logfile, secret_key, btc_amount, xch_amount_mojos):
         + ADDITIONAL_DATA
     )
 
-    total_sig = blspy.AugSchemeMPL.sign(private_key_for_secret(s), message)
+    private_key = private_key_for_secret(my_secret)
+    total_sig = blspy.AugSchemeMPL.sign(private_key, message)
     spend_bundle = SpendBundle(spend_bundle.coin_spends, total_sig)
 
     spend_bundle_hex = bytes(spend_bundle).hex()
     print(f"sweep spend bundle: {spend_bundle_hex}")
-    # spend_bundle.debug()
     print()
     print("Your counterparty should share their (disposable) private key")
     print("with you now. If your counterparty disappears before sending it,")
     print("you can use the spend bundle above as a last resort.")
     print()
+    return spend_bundle
 
-    remote_secret = ui_get_private_key(logfile, clawback_public_key)
-    total_secret = remote_secret + s + synthetic_offset
+
+def handle_remote_secret(
+    coin_spend,
+    message,
+    remote_secret,
+    my_secret,
+    synthetic_offset,
+    parent_coin_id,
+    xch_amount_mojos,
+    total_pubkey,
+    clawback_delay_seconds,
+    clawback_public_key,
+    sweep_receipt_hash,
+    sweep_public_key,
+    conditions,
+):
+    total_secret = remote_secret + my_secret + synthetic_offset
 
     total_private_key = private_key_for_secret(total_secret)
 
-    print()
-
     # build local signatures
 
     total_sig = blspy.AugSchemeMPL.sign(total_private_key, message)
     clean_spend_bundle = SpendBundle([coin_spend], total_sig)
-    # clean_spend_bundle.debug()
-    print(f"clean spend bundle: {bytes(clean_spend_bundle).hex()}")
+    spend_bundle_hex = bytes(clean_spend_bundle).hex()
+    print(f"clean spend bundle: {spend_bundle_hex}")
     print()
     print("Use the spend bundle above because it's smaller and is")
     print("indistinguishable from standard spend, so will give")
     print("the participants (you) more privacy.")
+    return clean_spend_bundle
 
 
 def main():
-    logfile, secret_key = ui_get_logfile()
-    btc_amount, xch_amount = ui_get_amounts(logfile)
+    with concurrent.futures.ThreadPoolExecutor(max_workers=5) as executor:
+        prices_future = executor.submit(lookup_xch_prices)
+        logfile, secret_key = ui_get_logfile()
+
+    prices = prices_future.result()
+    btc_amount, xch_amount = ui_get_amounts(logfile, prices)
     xch_amount_mojos = int(xch_amount * Decimal(1e12))
     which_way = ui_choose(logfile)
-    print()
+
     f = have_xch_want_btc if which_way == 1 else have_btc_want_xch
     f(logfile, secret_key, btc_amount, xch_amount_mojos)
 
